"""
- `spec_ops`: Provide several convolution of spectrum with a polynomial.

"""
from .spec_ops import conv_eqi_spec_with_slit, conv_spec_with_slit

__all__ = ["conv_eqi_spec_with_slit", "conv_spec_with_slit"]
